#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA resource utilization, etc.
"""
from a_sync import block
from pyramid.response import Response
from pyramid.view import view_config

from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.metrics import metastatus_lib


def parse_filters(filters):
    # The swagger config verifies that the data is in this format
    #  "pattern": "(.*):(.*,)*(.*)"
    if filters is None:
        return {}
    f = {s[0]: s[1] for s in [e.split(":") for e in filters]}
    f = {k: v.split(",") for k, v in f.items()}
    return f


@view_config(route_name="resources.utilization", request_method="GET", renderer="json")
def resources_utilization(request):
    master = get_mesos_master()
    mesos_state = block(master.state)

    groupings = request.swagger_data.get("groupings", ["superregion"])
    # swagger actually makes the key None if it's not set
    if groupings is None:
        groupings = ["superregion"]
    grouping_function = metastatus_lib.key_func_for_attribute_multi(groupings)
    sorting_function = metastatus_lib.sort_func_for_attributes(groupings)

    filters = request.swagger_data.get("filter", [])
    filters = parse_filters(filters)
    filter_funcs = [
        metastatus_lib.make_filter_slave_func(attr, vals)
        for attr, vals in filters.items()
    ]

    resource_info_dict = metastatus_lib.get_resource_utilization_by_grouping(
        grouping_func=grouping_function,
        mesos_state=mesos_state,
        filters=filter_funcs,
        sort_func=sorting_function,
    )

    response_body = []
    for k, v in resource_info_dict.items():
        group = {"groupings": {}}
        for grouping, value in k:
            group["groupings"][grouping] = value
        for resource, value in v["total"]._asdict().items():
            group[resource] = {"total": value}
        for resource, value in v["free"]._asdict().items():
            group[resource]["free"] = value
        for resource in v["free"]._fields:
            group[resource]["used"] = group[resource]["total"] - group[resource]["free"]

        response_body.append(group)

    return Response(json_body=response_body, status_code=200)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import socket
import sys
from typing import Dict
from typing import List

import choice

from paasta_tools import remote_git
from paasta_tools import utils
from paasta_tools.api.client import get_paasta_oapi_client
from paasta_tools.cli.cmds.mark_for_deployment import can_user_deploy_service
from paasta_tools.cli.cmds.mark_for_deployment import get_deploy_info
from paasta_tools.cli.cmds.status import add_instance_filter_arguments
from paasta_tools.cli.cmds.status import apply_args_filters
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.cli.utils import get_paasta_oapi_api_clustername
from paasta_tools.cli.utils import trigger_deploys
from paasta_tools.flink_tools import FlinkDeploymentConfig
from paasta_tools.flinkeks_tools import FlinkEksDeploymentConfig
from paasta_tools.generate_deployments_for_service import get_latest_deployment_tag
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PaastaColors


def add_subparser(subparsers):
    for command, lower, upper, cmd_func in [
        ("start", "start or restart", "Start or restart", paasta_start),
        ("restart", "start or restart", "Start or restart", paasta_restart),
        ("stop", "stop", "Stop", paasta_stop),
    ]:
        status_parser = subparsers.add_parser(
            command,
            help="%ss a PaaSTA service in a graceful way." % upper,
            description=(
                "%ss a PaaSTA service in a graceful way. This uses the Git control plane."
                % upper
            ),
            epilog=(
                "This command uses Git, and assumes access and authorization to the Git repo "
                "for the service is available."
            ),
        )
        add_instance_filter_arguments(status_parser, verb=lower)
        status_parser.add_argument(
            "-d",
            "--soa-dir",
            dest="soa_dir",
            metavar="SOA_DIR",
            default=DEFAULT_SOA_DIR,
            help="define a different soa config directory",
        )
        status_parser.set_defaults(command=cmd_func)


def format_tag(branch, force_bounce, desired_state):
    return f"refs/tags/paasta-{branch}-{force_bounce}-{desired_state}"


def make_mutate_refs_func(service_config, force_bounce, desired_state):
    """Create a function that will inform send_pack that we want to create tags
    corresponding to the set of branches passed, with the given force_bounce
    and desired_state parameters. These tags will point at the current tip of
    the branch they associate with.

    dulwich's send_pack wants a function that takes a dictionary of ref name
    to sha and returns a modified version of that dictionary. send_pack will
    then diff what is returned versus what was passed in, and inform the remote
    git repo of our desires."""

    def mutate_refs(refs):
        deploy_group = service_config.get_deploy_group()
        (_, head_sha, _) = get_latest_deployment_tag(refs, deploy_group)
        refs[
            format_tag(service_config.get_branch(), force_bounce, desired_state)
        ] = head_sha
        return refs

    return mutate_refs


def log_event(service_config, desired_state):
    user = utils.get_username()
    host = socket.getfqdn()
    line = "Issued request to change state of {} (an instance of {}) to '{}' by {}@{}".format(
        service_config.get_instance(),
        service_config.get_service(),
        desired_state,
        user,
        host,
    )
    utils._log(
        service=service_config.get_service(),
        level="event",
        cluster=service_config.get_cluster(),
        instance=service_config.get_instance(),
        component="deploy",
        line=line,
    )

    utils._log_audit(
        action=desired_state,
        service=service_config.get_service(),
        cluster=service_config.get_cluster(),
        instance=service_config.get_instance(),
    )


def issue_state_change_for_service(service_config, force_bounce, desired_state):
    ref_mutator = make_mutate_refs_func(
        service_config=service_config,
        force_bounce=force_bounce,
        desired_state=desired_state,
    )
    git_url = utils.get_git_url(service_config.get_service())
    remote_git.create_remote_refs(git_url, ref_mutator)
    if "yelpcorp.com" in git_url:
        trigger_deploys(service_config.get_service())
    log_event(service_config=service_config, desired_state=desired_state)


def print_kubernetes_message(desired_state):
    if desired_state == "start":
        print(
            "This service will soon be gracefully started/restarted, replacing old instances according "
            "to the bounce method chosen in soa-configs. "
        )
    elif desired_state == "stop":
        print(
            "This service will be gracefully stopped soon. It will be started back up again on the next deploy.\n"
            "To stop this service permanently. Set this in the soa-configs definition:\n"
            "\n"
            "    instances: 0\n"
        )


def print_flink_message(desired_state):
    if desired_state == "start":
        print("'Start' will tell Flink operator to start the cluster.")
    elif desired_state == "stop":
        print(
            "'Stop' will put Flink cluster in stopping mode, it may"
            "take some time before shutdown is completed."
        )


def confirm_to_continue(cluster_service_instances, desired_state):
    print(f"You are about to {desired_state} the following instances:")
    print("Either --instances or --clusters not specified. Asking for confirmation.")
    i_count = 0
    for cluster, services_instances in cluster_service_instances:
        for service, instances in services_instances.items():
            for instance in instances.keys():
                print(f"cluster = {cluster}, instance = {instance}")
                i_count += 1
    if sys.stdin.isatty():
        return choice.Binary(
            f"Are you sure you want to {desired_state} these {i_count} instances?",
            False,
        ).ask()
    return True


REMOTE_REFS: Dict[str, List[str]] = {}


def get_remote_refs(service, soa_dir):
    if service not in REMOTE_REFS:
        REMOTE_REFS[service] = remote_git.list_remote_refs(
            utils.get_git_url(service, soa_dir)
        )
    return REMOTE_REFS[service]


def paasta_start_or_stop(args, desired_state):
    """Requests a change of state to start or stop given branches of a service."""
    soa_dir = args.soa_dir

    pargs = apply_args_filters(args)
    if len(pargs) == 0:
        return 1

    affected_services = {
        s for service_list in pargs.values() for s in service_list.keys()
    }
    if len(affected_services) > 1:
        print(
            PaastaColors.red("Warning: trying to start/stop/restart multiple services:")
        )

        for cluster, services_instances in pargs.items():
            print("Cluster %s:" % cluster)
            for service, instances in services_instances.items():
                print("    Service %s:" % service)
                print("        Instances %s" % ",".join(instances.keys()))

        if sys.stdin.isatty():
            confirm = choice.Binary("Are you sure you want to continue?", False).ask()
        else:
            confirm = False
        if not confirm:
            print()
            print("exiting")
            return 1

    if not all(
        [
            can_user_deploy_service(get_deploy_info(service, soa_dir), service)
            for service in affected_services
        ]
    ):
        print(PaastaColors.red("Exiting due to missing deploy permissions"))
        return 1

    invalid_deploy_groups = []
    kubernetes_message_printed = False
    affected_flinks = []

    if args.clusters is None or args.instances is None:
        if confirm_to_continue(pargs.items(), desired_state) is False:
            print()
            print("exiting")
            return 1

    for cluster, services_instances in pargs.items():
        for service, instances in services_instances.items():
            for instance in instances.keys():
                service_config = get_instance_config(
                    service=service,
                    cluster=cluster,
                    instance=instance,
                    soa_dir=soa_dir,
                    load_deployments=False,
                )
                if isinstance(service_config, FlinkDeploymentConfig):
                    affected_flinks.append(service_config)
                    continue

                try:
                    remote_refs = get_remote_refs(service, soa_dir)
                except remote_git.LSRemoteException as e:
                    msg = (
                        "Error talking to the git server: %s\n"
                        "This PaaSTA command requires access to the git server to operate.\n"
                        "The git server may be down or not reachable from here.\n"
                        "Try again from somewhere where the git server can be reached, "
                        "like your developer environment."
                    ) % str(e)
                    print(msg)
                    return 1

                deploy_group = service_config.get_deploy_group()
                (deploy_tag, _, _) = get_latest_deployment_tag(
                    remote_refs, deploy_group
                )

                if deploy_tag not in remote_refs:
                    invalid_deploy_groups.append(deploy_group)
                else:
                    force_bounce = utils.format_timestamp(datetime.datetime.utcnow())
                    if (
                        isinstance(service_config, KubernetesDeploymentConfig)
                        and not kubernetes_message_printed
                    ):
                        print_kubernetes_message(desired_state)
                        kubernetes_message_printed = True

                    issue_state_change_for_service(
                        service_config=service_config,
                        force_bounce=force_bounce,
                        desired_state=desired_state,
                    )

    return_val = 0

    # TODO: Refactor to discover if set_state is available for given
    #       instance_type in API
    if affected_flinks:
        print_flink_message(desired_state)

        system_paasta_config = load_system_paasta_config()
        for service_config in affected_flinks:
            cluster = service_config.cluster
            service = service_config.service
            instance = service_config.instance
            is_eks = isinstance(service_config, FlinkEksDeploymentConfig)

            client = get_paasta_oapi_client(
                cluster=get_paasta_oapi_api_clustername(cluster=cluster, is_eks=is_eks),
                system_paasta_config=system_paasta_config,
            )
            if not client:
                print("Cannot get a paasta-api client")
                exit(1)

            try:
                client.service.instance_set_state(
                    service=service,
                    instance=instance,
                    desired_state=desired_state,
                )
            except client.api_error as exc:
                print(exc.reason)
                return exc.status

            return_val = 0

    if invalid_deploy_groups:
        print(f"No deploy tags found for {', '.join(invalid_deploy_groups)}.")
        print(f"Has {service} been deployed there yet?")
        return_val = 1

    return return_val


def paasta_start(args):
    return paasta_start_or_stop(args, "start")


def paasta_restart(args):
    pargs = apply_args_filters(args)
    soa_dir = args.soa_dir

    affected_flinks = []
    affected_non_flinks = []
    for cluster, service_instances in pargs.items():
        for service, instances in service_instances.items():
            for instance in instances.keys():
                service_config = get_instance_config(
                    service=service,
                    cluster=cluster,
                    instance=instance,
                    soa_dir=soa_dir,
                    load_deployments=False,
                )
                if isinstance(service_config, FlinkDeploymentConfig):
                    affected_flinks.append(service_config)
                else:
                    affected_non_flinks.append(service_config)

    if affected_flinks:
        flinks_info = ", ".join([f"{f.service}.{f.instance}" for f in affected_flinks])
        print(
            f"WARN: paasta restart is currently unsupported for Flink instances ({flinks_info})."
        )
        print("To restart, please run:", end="\n\n")
        for flink in affected_flinks:
            print(
                f"paasta stop -s {flink.service} -i {flink.instance} -c {flink.cluster}"
            )
            print(
                f"paasta start -s {flink.service} -i {flink.instance} -c {flink.cluster}",
                end="\n\n",
            )

        if not affected_non_flinks:
            return 1

        non_flinks_info = ", ".join(
            [f"{f.service}.{f.instance}" for f in affected_non_flinks]
        )
        proceed = choice.Binary(
            f"Would you like to restart the other instances ({non_flinks_info}) anyway?",
            False,
        ).ask()

        if not proceed:
            return 1

    return paasta_start(args)


PAASTA_STOP_UNDERSPECIFIED_ARGS_MESSAGE = PaastaColors.red(
    "paasta stop requires explicit specification of cluster, service, and instance."
)


def paasta_stop(args):
    if not args.clusters:
        print(PAASTA_STOP_UNDERSPECIFIED_ARGS_MESSAGE)
        return 1
    elif not args.service_instance and not (args.service and args.instances):
        print(PAASTA_STOP_UNDERSPECIFIED_ARGS_MESSAGE)
        return 1
    return paasta_start_or_stop(args, "stop")



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_crd.py <service.crd> [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -c <cluster>, --cluster <cluster>: Specify a kubernetes cluster name
- -v, --verbose: Verbose output
"""
import argparse
import logging
import sys
from typing import Sequence

import service_configuration_lib
from kubernetes.client import V1beta1CustomResourceDefinition
from kubernetes.client import V1CustomResourceDefinition
from kubernetes.client.exceptions import ApiException

from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import paasta_prefixed
from paasta_tools.kubernetes_tools import update_crds
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates/updates kubernetes CRDs.")
    parser.add_argument(
        "service_list",
        nargs="+",
        help="The list of services to create or update CRDs for",
        metavar="SERVICE",
    )
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        metavar="CLUSTER",
        default=None,
        help="Kubernetes cluster name",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    if args.cluster:
        cluster = args.cluster
    else:
        system_paasta_config = load_system_paasta_config()
        cluster = system_paasta_config.get_cluster()

    kube_client = KubeClient()

    success = setup_kube_crd(
        kube_client=kube_client,
        cluster=cluster,
        services=args.service_list,
        soa_dir=soa_dir,
    )
    sys.exit(0 if success else 1)


def setup_kube_crd(
    kube_client: KubeClient,
    cluster: str,
    services: Sequence[str],
    soa_dir: str = DEFAULT_SOA_DIR,
) -> bool:
    existing_crds = kube_client.apiextensions.list_custom_resource_definition(
        label_selector=paasta_prefixed("service")
    )

    # This step can fail in k8s 1.22 since this version is not existing anymore
    # we need to support this for the transition
    try:
        existing_crds_v1_beta1 = (
            kube_client.apiextensions_v1_beta1.list_custom_resource_definition(
                label_selector=paasta_prefixed("service")
            )
        )
    except ApiException:
        existing_crds_v1_beta1 = []
        log.debug(
            "Listing CRDs with apiextensions/v1beta1 not supported on this cluster, falling back to v1"
        )

    desired_crds = []
    desired_crds_v1_beta1 = []
    for service in services:
        crd_config = service_configuration_lib.read_extra_service_information(
            service, f"crd-{cluster}", soa_dir=soa_dir
        )
        if not crd_config:
            log.info("nothing to deploy")
            continue

        metadata = crd_config.get("metadata", {})
        if "labels" not in metadata:
            metadata["labels"] = {}
        metadata["labels"]["yelp.com/paasta_service"] = service
        metadata["labels"][paasta_prefixed("service")] = service

        if "apiextensions.k8s.io/v1beta1" == crd_config["apiVersion"]:
            desired_crd = V1beta1CustomResourceDefinition(
                api_version=crd_config.get("apiVersion"),
                kind=crd_config.get("kind"),
                metadata=metadata,
                spec=crd_config.get("spec"),
            )
            desired_crds_v1_beta1.append(desired_crd)
        else:
            desired_crd = V1CustomResourceDefinition(
                api_version=crd_config.get("apiVersion"),
                kind=crd_config.get("kind"),
                metadata=metadata,
                spec=crd_config.get("spec"),
            )
            desired_crds.append(desired_crd)

    return update_crds(
        kube_client=kube_client,
        desired_crds=desired_crds,
        existing_crds=existing_crds,
    ) and update_crds(
        kube_client=kube_client,
        desired_crds=desired_crds_v1_beta1,
        existing_crds=existing_crds_v1_beta1,
    )


if __name__ == "__main__":
    main()



# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import logging
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type

from service_configuration_lib import read_service_configuration

from paasta_tools import utils
from paasta_tools.autoscaling.utils import AutoscalingParamsDict
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig_T
from paasta_tools.utils import list_clusters
from paasta_tools.utils import load_service_instance_configs
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import NoDeploymentsAvailable


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class PaastaServiceConfigLoader:
    """PaastaServiceConfigLoader provides useful methods for reading soa-configs and
    iterating instance names or InstanceConfigs objects.

    :Example:

    >>> from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
    >>> from paasta_tools.utils import DEFAULT_SOA_DIR
    >>>
    >>> sc = PaastaServiceConfigLoader(service='fake_service', soa_dir=DEFAULT_SOA_DIR)
    >>>
    >>> for instance in sc.instances(cluster='fake_cluster', instance_type_class=KubernetesDeploymentConfig):
    ...     print(instance)
    ...
    main
    canary
    >>>
    >>> for instance_config in sc.instance_configs(cluster='fake_cluster', instance_type_class=KubernetesDeploymentConfig):
    ...     print(instance_config.get_instance())
    ...
    main
    canary
    >>>
    """

    _framework_configs: Dict[Tuple[str, type], Dict[str, utils.InstanceConfigDict]]
    _clusters: List[str]
    _deployments_json: utils.DeploymentsJsonV2

    def __init__(
        self,
        service: str,
        soa_dir: str = DEFAULT_SOA_DIR,
        load_deployments: bool = True,
    ) -> None:
        self._service = service
        self._soa_dir = soa_dir
        self._load_deployments = load_deployments
        self._clusters = None
        self._general_config = None
        self._deployments_json = None
        self._framework_configs = {}

    @property
    def clusters(self) -> Iterable[str]:
        """Returns an iterator that yields cluster names for the service.

        :returns: iterator that yields cluster names.
        """
        if self._clusters is None:
            self._clusters = list_clusters(service=self._service, soa_dir=self._soa_dir)
        for cluster in self._clusters:
            yield cluster

    def instances(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ) -> Iterable[str]:
        """Returns an iterator that yields instance names as strings.

        :param cluster: The cluster name
        :param instance_type_class: a subclass of InstanceConfig
        :returns: an iterator that yields instance names
        """
        if (cluster, instance_type_class) not in self._framework_configs:
            self._refresh_framework_config(cluster, instance_type_class)
        for instance in self._framework_configs.get((cluster, instance_type_class), []):
            yield instance

    def instance_configs(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ) -> Iterable[InstanceConfig_T]:
        """Returns an iterator that yields InstanceConfig objects.

        :param cluster: The cluster name
        :param instance_type_class: a subclass of InstanceConfig
        :returns: an iterator that yields instances of KubernetesDeploymentConfig, etc.
        :raises NotImplementedError: when it doesn't know how to create a config for instance_type_class
        """
        if (cluster, instance_type_class) not in self._framework_configs:
            self._refresh_framework_config(cluster, instance_type_class)
        for instance, config in self._framework_configs.get(
            (cluster, instance_type_class), {}
        ).items():
            try:
                yield self._create_service_config(
                    cluster, instance, config, instance_type_class
                )
            except NoDeploymentsAvailable:
                pass

    def _framework_config_filename(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ):
        return f"{instance_type_class.config_filename_prefix}-{cluster}"

    def _refresh_framework_config(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ):
        instances = load_service_instance_configs(
            service=self._service,
            instance_type=instance_type_class.config_filename_prefix,
            cluster=cluster,
            soa_dir=self._soa_dir,
        )
        self._framework_configs[(cluster, instance_type_class)] = instances

    def _get_branch_dict(
        self, cluster: str, instance: str, config: utils.InstanceConfig
    ) -> utils.BranchDictV2:
        if self._deployments_json is None:
            self._deployments_json = load_v2_deployments_json(
                self._service, soa_dir=self._soa_dir
            )

        branch = config.get_branch()
        deploy_group = config.get_deploy_group()
        return self._deployments_json.get_branch_dict(
            self._service, branch, deploy_group
        )

    def _get_merged_config(
        self, config: utils.InstanceConfigDict
    ) -> utils.InstanceConfigDict:
        if self._general_config is None:
            self._general_config = read_service_configuration(
                service_name=self._service, soa_dir=self._soa_dir
            )
        return deep_merge_dictionaries(overrides=config, defaults=self._general_config)

    def _create_service_config(
        self,
        cluster: str,
        instance: str,
        config: utils.InstanceConfigDict,
        config_class: Type[InstanceConfig_T],
    ) -> InstanceConfig_T:
        """Create a service instance's configuration for kubernetes.

        :param cluster: The cluster to read the configuration for
        :param instance: The instance of the service to retrieve
        :param config: the framework instance config.
        :returns: An instance of config_class
        """

        merged_config = self._get_merged_config(config)

        # These type: ignore annotations will go away once yelpsoa-configs is migrated (COREJAVA-1339)
        if (
            "autoscaling" in merged_config
            and "metrics_providers" not in merged_config["autoscaling"]  # type: ignore
        ):
            merged_config["autoscaling"] = transform_autoscaling_params_dict(  # type: ignore
                copy.deepcopy(merged_config["autoscaling"])  # type: ignore
            )

        temp_instance_config = config_class(
            service=self._service,
            cluster=cluster,
            instance=instance,
            config_dict=merged_config,
            branch_dict=None,
            soa_dir=self._soa_dir,
        )

        branch_dict = self._get_branch_dict(cluster, instance, temp_instance_config)

        return config_class(
            service=self._service,
            cluster=cluster,
            instance=instance,
            config_dict=merged_config,
            branch_dict=branch_dict,
            soa_dir=self._soa_dir,
        )


# Remove this once yelpsoa-configs is using the new format (COREJAVA-1339)
def transform_autoscaling_params_dict(
    old_autoscaling_params: Dict[str, Any]
) -> AutoscalingParamsDict:
    metrics_provider_type = old_autoscaling_params.pop("metrics_provider", "cpu")
    old_autoscaling_params["type"] = metrics_provider_type
    scaledown_policies = old_autoscaling_params.pop("scaledown_policies", None)

    new_autoscaling_params = {"metrics_providers": [old_autoscaling_params]}
    if scaledown_policies is not None:
        new_autoscaling_params["scaledown_policies"] = scaledown_policies

    return new_autoscaling_params  # type: ignore



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Responds to paasta service and instance requests.
"""
import argparse
import contextlib
import logging
import os
import sys

import manhole
import requests_cache
import service_configuration_lib
import yaml
from pyramid.config import Configurator
from wsgicors import CORS

import paasta_tools.api
from paasta_tools import kubernetes_tools
from paasta_tools.api import settings
from paasta_tools.api.tweens import profiling
from paasta_tools.api.tweens import request_logger
from paasta_tools.utils import load_system_paasta_config

try:
    import clog
except ImportError:
    clog = None


log = logging.getLogger(__name__)


def parse_paasta_api_args():
    parser = argparse.ArgumentParser(description="Runs a PaaSTA API server")
    parser.add_argument(
        "-D",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="output the debug logs",
    )
    parser.add_argument("port", type=int, help="port number for the api server")
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        help="specify a cluster. If no empty, the cluster from /etc/paasta is used",
    )
    parser.add_argument(
        "--max-request-seconds",
        default=120,
        dest="max_request_seconds",
        help="Maximum seconds allowed for a worker to process a request",
    )
    parser.add_argument(
        "-w",
        "--workers",
        default=4,
        help="Number of gunicorn workers to run",
    )
    args = parser.parse_args()
    return args


def make_app(global_config=None):
    paasta_api_path = os.path.dirname(paasta_tools.api.__file__)
    setup_paasta_api()
    setup_clog()

    config = Configurator(
        settings={
            "service_name": "paasta-api",
            "pyramid_swagger.schema_directory": os.path.join(
                paasta_api_path, "api_docs"
            ),
            "pyramid_swagger.skip_validation": [
                "/(static)\\b",
                "/(status)\\b",
                "/(swagger.json)\\b",
            ],
            "pyramid_swagger.swagger_versions": ["2.0"],
        }
    )

    config.include("pyramid_swagger")
    config.include(request_logger)

    config.add_route(
        "flink.service.instance.jobs", "/v1/flink/{service}/{instance}/jobs"
    )

    config.add_route(
        "flink.service.instance.job_details",
        "/v1/flink/{service}/{instance}/jobs/{job_id}",
    )

    config.add_route(
        "flink.service.instance.overview", "/v1/flink/{service}/{instance}/overview"
    )
    config.add_route(
        "flink.service.instance.config", "/v1/flink/{service}/{instance}/config"
    )
    config.include(profiling)

    config.add_route("resources.utilization", "/v1/resources/utilization")
    config.add_route(
        "service.instance.status", "/v1/services/{service}/{instance}/status"
    )
    config.add_route(
        "service.instance.mesh_status", "/v1/services/{service}/{instance}/mesh_status"
    )
    config.add_route(
        "service.instance.bounce_status",
        "/v1/services/{service}/{instance}/bounce_status",
    )
    config.add_route(
        "service.instance.set_state",
        "/v1/services/{service}/{instance}/state/{desired_state}",
    )
    config.add_route(
        "service.instance.delay", "/v1/services/{service}/{instance}/delay"
    )
    config.add_route(
        "service.instance.tasks", "/v1/services/{service}/{instance}/tasks"
    )
    config.add_route(
        "service.instance.tasks.task",
        "/v1/services/{service}/{instance}/tasks/{task_id}",
    )
    config.add_route("service.list", "/v1/services/{service}")
    config.add_route("services", "/v1/services")
    config.add_route(
        "service.autoscaler.get",
        "/v1/services/{service}/{instance}/autoscaler",
        request_method="GET",
    )
    config.add_route(
        "service.autoscaler.post",
        "/v1/services/{service}/{instance}/autoscaler",
        request_method="POST",
    )
    config.add_route(
        "service_autoscaler.pause.post",
        "/v1/service_autoscaler/pause",
        request_method="POST",
    )
    config.add_route(
        "service_autoscaler.pause.delete",
        "/v1/service_autoscaler/pause",
        request_method="DELETE",
    )
    config.add_route(
        "service_autoscaler.pause.get",
        "/v1/service_autoscaler/pause",
        request_method="GET",
    )
    config.add_route("version", "/v1/version")
    config.add_route("deploy_queue.list", "/v1/deploy_queue")
    config.scan()
    return CORS(
        config.make_wsgi_app(), headers="*", methods="*", maxage="180", origin="*"
    )


_app = None


def application(env, start_response):
    """For uwsgi or gunicorn."""
    global _app
    if not _app:
        _app = make_app()
        manhole_path = os.environ.get("PAASTA_MANHOLE_PATH")
        if manhole_path:
            manhole.install(
                socket_path=f"{manhole_path}-{os.getpid()}", locals={"_app": _app}
            )
    return _app(env, start_response)


def setup_paasta_api():
    if os.environ.get("PAASTA_API_DEBUG"):
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    # pyinotify is a better solution than turning off file caching completely
    service_configuration_lib.disable_yaml_cache()

    settings.system_paasta_config = load_system_paasta_config()
    if os.environ.get("PAASTA_API_CLUSTER"):
        settings.cluster = os.environ.get("PAASTA_API_CLUSTER")
    else:
        settings.cluster = settings.system_paasta_config.get_cluster()

    try:
        settings.kubernetes_client = kubernetes_tools.KubeClient()
    except FileNotFoundError:
        log.info("Kubernetes not found")
        settings.kubernetes_client = None
    except Exception:
        log.exception("Error while initializing KubeClient")
        settings.kubernetes_client = None

    # Set up transparent cache for http API calls. With expire_after, responses
    # are removed only when the same request is made. Expired storage is not a
    # concern here. Thus remove_expired_responses is not needed.
    requests_cache.install_cache("paasta-api", backend="memory", expire_after=5)


def setup_clog(config_file="/nail/srv/configs/clog.yaml"):
    if clog:
        if os.path.exists(config_file):
            with open(config_file) as fp:
                clog_config = yaml.safe_load(fp)
        else:
            # these are barebones basic configs from /nail/srv/configs/clog.yaml
            clog_config = {
                "scribe_host": "169.254.255.254",
                "scribe_port": 1463,
                "monk_disable": False,
                "scribe_disable": False,
            }
        clog.config.configure_from_dict(clog_config)


def main(argv=None):
    args = parse_paasta_api_args()

    if args.debug:
        os.environ["PAASTA_API_DEBUG"] = "1"

    if args.soa_dir:
        os.environ["PAASTA_API_SOA_DIR"] = args.soa_dir

    if args.cluster:
        os.environ["PAASTA_API_CLUSTER"] = args.cluster

    gunicorn_args = [
        "gunicorn",
        "-w",
        str(args.workers),
        "--bind",
        f":{args.port}",
        "--timeout",
        str(args.max_request_seconds),
        "--graceful-timeout",
        str(args.max_request_seconds),
        "paasta_tools.api.api:application",
    ]

    if argv:
        with redirect_argv(gunicorn_args):
            from gunicorn.app import wsgiapp

            wsgiapp.run()
    else:
        os.execlp(os.path.join(sys.exec_prefix, "bin", "gunicorn"), *gunicorn_args)


@contextlib.contextmanager
def redirect_argv(args):
    sys._argv = sys.argv[:]
    sys.argv = args
    yield
    sys.argv = sys._argv


if __name__ == "__main__":
    main()
